# Copyright 2023 The Navix Authors.

# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.


from __future__ import annotations

from typing import Callable, Tuple
from jax import Array
import jax
import jax.numpy as jnp
import jax.tree_util as jtu
from .entities import Entities, Ball
from .states import EventsManager, State
from .grid import positions_equal, translate


def deterministic_transition(
    state: State, action: Array, actions_set: Tuple[Callable[[State], State], ...]
) -> State:
    return jax.lax.switch(action, actions_set, state)


def stochastic_transition(
    state: State, action: Array, actions_set: Tuple[Callable[[State], State], ...]
) -> State:
    # actions
    state = jax.lax.switch(action, actions_set, state)

    state = update_balls(state)
    return state


def update_balls(state: State) -> State:
    def update_one(position, key):
        direction = jax.random.randint(key, (), minval=0, maxval=4)
        new_position = translate(position, direction)
        can_move, events = _can_spawn_there(state, new_position)
        return jnp.where(can_move, new_position, position), events

    if Entities.BALL in state.entities:
        balls: Ball = state.entities[Entities.BALL]  # type: ignore
        keys = jax.random.split(state.key, len(balls.position) + 1)
        new_position, events = jax.jit(jax.vmap(update_one))(balls.position, keys[1:])
        # update structs
        balls = balls.replace(position=new_position)
        state = state.set_balls(balls)
        events = jtu.tree_map(lambda x: jnp.any(x), events)
        state = state.replace(key=keys[0], events=events)
    return state


def _can_spawn_there(state: State, position: Array) -> Tuple[Array, EventsManager]:
    # according to the grid
    walkable = jnp.equal(state.grid[tuple(position)], 0)

    # according to entities
    events = state.events
    for k in state.entities:
        obstructs = positions_equal(state.entities[k].position, position)
        if k == Entities.PLAYER:
            events = jax.lax.cond(
                jnp.any(obstructs),
                lambda x: x.record_ball_hit(state.entities[k], position),
                lambda x: x,
                events,
            )
        walkable = jnp.logical_and(walkable, jnp.any(jnp.logical_not(obstructs)))
    return jnp.asarray(walkable, dtype=jnp.bool_), events


DEFAULT_TRANSITION = stochastic_transition



from __future__ import annotations
from typing import Tuple, TypeVar

import jax
from jax import Array
import jax.numpy as jnp
from flax import struct


from .components import (
    Positionable,
    Directional,
    HasColour,
    HasTag,
    Stochastic,
    Openable,
    Pickable,
    Holder,
    HasSprite,
)
from .rendering.registry import SPRITES_REGISTRY

T = TypeVar("T", bound="Entity")


class Entities(struct.PyTreeNode):
    WALL: str = struct.field(pytree_node=False, default="wall")
    FLOOR: str = struct.field(pytree_node=False, default="floor")
    PLAYER: str = struct.field(pytree_node=False, default="player")
    GOAL: str = struct.field(pytree_node=False, default="goal")
    KEY: str = struct.field(pytree_node=False, default="key")
    DOOR: str = struct.field(pytree_node=False, default="door")
    LAVA: str = struct.field(pytree_node=False, default="lava")
    BALL: str = struct.field(pytree_node=False, default="ball")
    BOX: str = struct.field(pytree_node=False, default="box")


class Entity(Positionable, HasTag, HasSprite):
    """Entities are components that can be placed in the environment"""

    # def __post_init__(self) -> None:
    #     if not config.ARRAY_CHECKS_ENABLED:
    #         return
    #     # Check that all fields have the same batch size
    #     fields = self.__dataclass_fields__
    #     batch_size = self.shape[0:]
    #     for path, leaf in jax.tree_util.tree_leaves_with_path(self):
    #         name = path[0].name
    #         default_ndim = len(fields[name].metadata["shape"])
    #         prefix = int(default_ndim != leaf.ndim)
    #         leaf_batch_size = leaf.shape[:prefix]
    #         assert (
    #             leaf_batch_size == batch_size
    #         ), f"Expected {name} to have batch size {batch_size}, got {leaf_batch_size} instead"

    # def check_ndim(self, batched: bool = False) -> None:
    #     if not config.ARRAY_CHECKS_ENABLED:
    #         return
    #     for field in dataclasses.fields(self):
    #         value = getattr(self, field.name)
    #         default_ndim = len(field.metadata["shape"])
    #         assert (
    #             value.ndim == default_ndim + batched
    #         ), f"Expected {field.name} to have ndim {default_ndim - batched}, got {value.ndim} instead"

    def __getitem__(self: T, idx) -> T:
        return jax.tree_util.tree_map(lambda x: x[idx], self)

    @property
    def name(self) -> str:
        return self.__class__.__name__

    @property
    def shape(self) -> Tuple[int, ...]:
        """The batch shape of the entity"""
        return self.position.shape[:-1]

    @property
    def ndim(self) -> int:
        return self.position.ndim - 1

    @property
    def walkable(self) -> Array:
        raise NotImplementedError()

    @property
    def transparent(self) -> Array:
        raise NotImplementedError()


class Wall(Entity):
    """Walls are entities that cannot be walked through"""

    @classmethod
    def create(
        cls,
        position: Array,
    ) -> Wall:
        return cls(position=position)

    @property
    def walkable(self) -> Array:
        return jnp.broadcast_to(jnp.asarray(False), self.shape)

    @property
    def transparent(self) -> Array:
        return jnp.broadcast_to(jnp.asarray(False), self.shape)

    @property
    def sprite(self) -> Array:
        sprite = SPRITES_REGISTRY[Entities.WALL]
        return jnp.broadcast_to(sprite[None], (*self.shape, *sprite.shape))

    @property
    def tag(self) -> Array:
        return jnp.broadcast_to(jnp.asarray(0), self.shape)


class Player(Entity, Directional, Holder):
    """Players are entities that can act around the environment"""

    @classmethod
    def create(
        cls,
        position: Array,
        direction: Array,
        pocket: Array,
    ) -> Player:
        return cls(position=position, direction=direction, pocket=pocket)

    @property
    def walkable(self) -> Array:
        return jnp.broadcast_to(jnp.asarray(True), self.shape)

    @property
    def transparent(self) -> Array:
        return jnp.broadcast_to(jnp.asarray(True), self.shape)

    @property
    def sprite(self) -> Array:
        sprite = SPRITES_REGISTRY[Entities.PLAYER][self.direction]
        if sprite.ndim == 3:
            # batch it
            sprite = sprite[None]
        # broadcast to batch_size
        return jnp.broadcast_to(sprite, (*self.shape, *sprite.shape[1:]))

    @property
    def tag(self) -> Array:
        return jnp.broadcast_to(jnp.asarray(2), self.shape)


class Goal(Entity, Stochastic):
    """Goals are entities that can be reached by the player"""

    @classmethod
    def create(
        cls,
        position: Array,
        probability: Array,
    ) -> Goal:
        return cls(position=position, probability=probability)

    @property
    def walkable(self) -> Array:
        return jnp.broadcast_to(jnp.asarray(True), self.shape)

    @property
    def transparent(self) -> Array:
        return jnp.broadcast_to(jnp.asarray(True), self.shape)

    @property
    def sprite(self) -> Array:
        sprite = SPRITES_REGISTRY[Entities.GOAL]
        if sprite.ndim == 3:
            # batch it
            sprite = sprite[None]
        # ensure same batch size
        if sprite.shape[0] != self.position.shape[0]:
            sprite = jnp.broadcast_to(sprite, (*self.shape, *sprite.shape[1:]))
        return sprite

    @property
    def tag(self) -> Array:
        return jnp.broadcast_to(jnp.asarray(3), self.shape)


class Key(Entity, Pickable, HasColour):
    """Pickable items are world objects that can be picked up by the player.
    Examples of pickable items are keys, coins, etc."""

    @classmethod
    def create(
        cls,
        position: Array,
        colour: Array,
        id: Array,
    ) -> Key:
        return cls(position=position, id=id, colour=colour)

    @property
    def walkable(self) -> Array:
        return jnp.broadcast_to(jnp.asarray(False), self.shape)

    @property
    def transparent(self) -> Array:
        return jnp.broadcast_to(jnp.asarray(True), self.shape)

    @property
    def sprite(self) -> Array:
        sprite = SPRITES_REGISTRY[Entities.KEY][self.colour]
        if sprite.ndim == 3:
            # batch it
            sprite = sprite[None]
        # ensure same batch size
        if sprite.shape[0] != self.position.shape[0]:
            sprite = jnp.broadcast_to(sprite, (*self.shape, *sprite.shape[1:]))
        return sprite

    @property
    def tag(self) -> Array:
        return jnp.broadcast_to(jnp.asarray(4), self.shape)


class Door(Entity, Openable, HasColour):
    """Consumable items are world objects that can be consumed by the player.
    Consuming an item requires a tool (e.g. a key to open a door).
    A tool is an id (int) of another item, specified in the `requires` field (-1 if no tool is required).
    After an item is consumed, it is both removed from the `state.entities` collection, and replaced in the grid
    by the item specified in the `replacement` field (0 = floor by default).
    Examples of consumables are doors (to open) food (to eat) and water (to drink), etc.
    """

    @classmethod
    def create(
        cls,
        position: Array,
        requires: Array,
        colour: Array,
        open: Array,
    ) -> Door:
        return cls(
            position=position,
            requires=requires,
            open=open,
            colour=colour,
        )

    @property
    def walkable(self) -> Array:
        return self.open

    @property
    def transparent(self) -> Array:
        return self.open

    @property
    def sprite(self) -> Array:
        sprite = SPRITES_REGISTRY[Entities.DOOR][
            self.colour, jnp.asarray(self.open + 2 * self.locked, dtype=jnp.int32)
        ]
        if sprite.ndim == 3:
            # batch it
            sprite = sprite[None]
        # ensure same batch size
        if sprite.shape[0] != self.position.shape[0]:
            sprite = jnp.broadcast_to(sprite, (*self.shape, *sprite.shape[1:]))
        return sprite

    @property
    def tag(self) -> Array:
        return jnp.broadcast_to(jnp.asarray(5), self.shape)

    @property
    def locked(self) -> Array:
        return self.requires != jnp.asarray(-1)


class Lava(Entity):
    """Goals are entities that can be reached by the player"""

    @classmethod
    def create(
        cls,
        position: Array,
    ) -> Lava:
        return cls(position=position)

    @property
    def walkable(self) -> Array:
        return jnp.broadcast_to(jnp.asarray(True), self.shape)

    @property
    def transparent(self) -> Array:
        return jnp.broadcast_to(jnp.asarray(True), self.shape)

    @property
    def sprite(self) -> Array:
        sprite = SPRITES_REGISTRY[Entities.LAVA]
        if sprite.ndim == 3:
            # batch it
            sprite = sprite[None]
        # ensure same batch size
        if sprite.shape[0] != self.position.shape[0]:
            sprite = jnp.broadcast_to(sprite, (*self.shape, *sprite.shape[1:]))
        return sprite

    @property
    def tag(self) -> Array:
        return jnp.broadcast_to(jnp.asarray(6), self.shape)


class Ball(Entity, HasColour, Stochastic):
    """Goals are entities that can be reached by the player"""

    @classmethod
    def create(
        cls,
        position: Array,
        colour: Array,
        probability: Array,
    ) -> Ball:
        return cls(position=position, colour=colour, probability=probability)

    @property
    def walkable(self) -> Array:
        return jnp.broadcast_to(jnp.asarray(False), self.shape)

    @property
    def transparent(self) -> Array:
        return jnp.broadcast_to(jnp.asarray(True), self.shape)

    @property
    def sprite(self) -> Array:
        sprite = SPRITES_REGISTRY[Entities.BALL][self.colour]
        if sprite.ndim == 3:
            # batch it
            sprite = sprite[None]
        # ensure same batch size
        if sprite.shape[0] != self.position.shape[0]:
            sprite = jnp.broadcast_to(sprite, (*self.shape, *sprite.shape[1:]))
        return sprite

    @property
    def tag(self) -> Array:
        return jnp.broadcast_to(jnp.asarray(7), self.shape)


class Box(Entity, HasColour, Holder):
    """Goals are entities that can be reached by the player"""

    @classmethod
    def create(
        cls,
        position: Array,
        colour: Array,
        pocket: Array,
    ) -> Box:
        return cls(position=position, colour=colour, pocket=pocket)

    @property
    def walkable(self) -> Array:
        return jnp.broadcast_to(jnp.asarray(False), self.shape)

    @property
    def transparent(self) -> Array:
        return jnp.broadcast_to(jnp.asarray(True), self.shape)

    @property
    def sprite(self) -> Array:
        sprite = SPRITES_REGISTRY[Entities.BOX][self.colour]
        if sprite.ndim == 3:
            # batch it
            sprite = sprite[None]
        # ensure same batch size
        if sprite.shape[0] != self.position.shape[0]:
            sprite = jnp.broadcast_to(sprite, (*self.shape, *sprite.shape[1:]))
        return sprite

    @property
    def tag(self) -> Array:
        return jnp.broadcast_to(jnp.asarray(8), self.shape)



# Copyright 2023 The Navix Authors.

# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations
from typing import Dict

from jax import Array
import jax.numpy as jnp
from flax import struct


from .components import (
    Positionable,
    HasColour,
)
from .rendering.cache import RenderingCache
from .entities import Entity, Entities, Goal, Wall, Ball, Lava, Key, Door, Box, Player


class EventType:
    NONE: Array = jnp.asarray(-1, dtype=jnp.int32)
    REACH: Array = jnp.asarray(0, dtype=jnp.int32)
    HIT: Array = jnp.asarray(1, dtype=jnp.int32)
    FALL: Array = jnp.asarray(2, dtype=jnp.int32)
    PICKUP: Array = jnp.asarray(3, dtype=jnp.int32)
    OPEN: Array = jnp.asarray(4, dtype=jnp.int32)
    UNLOCK: Array = jnp.asarray(5, dtype=jnp.int32)


class Event(Positionable, HasColour):
    position: Array = jnp.asarray([-1, -1], dtype=jnp.int32)
    colour: Array = jnp.asarray(-1, dtype=jnp.int32)
    happened: Array = jnp.asarray(False, dtype=jnp.bool_)
    event_type: Array = EventType.NONE

    def __eq__(self, other: Event) -> Array:
        return jnp.logical_and(
            jnp.array_equal(self.position, other.position),
            jnp.array_equal(self.colour, other.colour),
        )

    def __ne__(self, other: Event) -> Array:
        return jnp.logical_not(self == other)


class EventsManager(struct.PyTreeNode):
    goal_reached: Event = Event()
    ball_hit: Event = Event()
    wall_hit: Event = Event()
    lava_fall: Event = Event()
    key_pickup: Event = Event()
    door_opening: Event = Event()
    door_unlock: Event = Event()
    ball_pickup: Event = Event()

    def record_walk_into(self, entity: Entity, position: Array) -> EventsManager:
        if isinstance(entity, Goal):
            return self.record_goal_reached(entity, position)
        elif isinstance(entity, Wall):
            return self.record_wall_hit(entity, position)
        elif isinstance(entity, Lava):
            return self.record_lava_fall(entity, position)
        return self

    def record_pickup(self, entity: Entity, position: Array) -> EventsManager:
        if isinstance(entity, Key):
            return self.record_key_pickup(entity, position)
        elif isinstance(entity, Ball):
            return self.record_ball_pickup(entity, position)
        return self

    def record_goal_reached(self, goal: Goal, position: Array) -> EventsManager:
        idx = jnp.where(goal.position == position, size=1)[0][0]
        goal = goal[idx]
        return self.replace(
            goal_reached=Event(
                position=position,
                colour=jnp.asarray(-1, dtype=jnp.int32),
                happened=jnp.asarray(True, dtype=jnp.bool_),
                event_type=EventType.REACH,
            )
        )

    def record_ball_hit(self, ball: Ball, position: Array) -> EventsManager:
        idx = jnp.where(ball.position == position, size=1)[0][0]
        ball = ball[idx]
        return self.replace(
            ball_hit=Event(
                position=ball.position,
                colour=ball.colour,
                happened=jnp.asarray(True, dtype=jnp.bool_),
                event_type=EventType.HIT,
            )
        )

    def record_wall_hit(self, wall: Wall, position: Array) -> EventsManager:
        idx = jnp.where(wall.position == position, size=1)[0][0]
        wall = wall[idx]
        return self.replace(
            wall_hit=Event(
                position=wall.position,
                colour=jnp.asarray(-1, dtype=jnp.int32),
                happened=jnp.asarray(True, dtype=jnp.bool_),
                event_type=EventType.HIT,
            )
        )

    def record_grid_hit(self, position: Array) -> EventsManager:
        return self.replace(
            wall_hit=Event(
                position=position,
                colour=jnp.asarray(-1, dtype=jnp.int32),
                happened=jnp.asarray(True, dtype=jnp.bool_),
                event_type=EventType.HIT,
            )
        )

    def record_lava_fall(self, lava: Lava, position: Array) -> EventsManager:
        idx = jnp.where(lava.position == position, size=1)[0][0]
        lava = lava[idx]
        return self.replace(
            lava_fall=Event(
                position=lava.position,
                colour=jnp.asarray(-1, dtype=jnp.int32),
                happened=jnp.asarray(True, dtype=jnp.bool_),
                event_type=EventType.FALL,
            )
        )

    def record_key_pickup(self, key: Key, position: Array) -> EventsManager:
        idx = jnp.where(key.position == position, size=1)[0][0]
        key = key[idx]
        return self.replace(
            key_pickup=Event(
                position=key.position,
                colour=key.colour,
                happened=jnp.asarray(True, dtype=jnp.bool_),
                event_type=EventType.PICKUP,
            )
        )

    def record_door_opening(self, door: Door, position: Array) -> EventsManager:
        idx = jnp.where(door.position == position, size=1)[0][0]
        door = door[idx]
        return self.replace(
            door_opening=Event(
                position=door.position,
                colour=door.colour,
                happened=jnp.asarray(True, dtype=jnp.bool_),
                event_type=EventType.OPEN,
            )
        )

    def record_door_unlock(self, door: Door, position: Array) -> EventsManager:
        idx = jnp.where(door.position == position, size=1)[0][0]
        door = door[idx]
        return self.replace(
            door_opening=Event(
                position=door.position,
                colour=door.colour,
                happened=jnp.asarray(True, dtype=jnp.bool_),
                event_type=EventType.UNLOCK,
            )
        )

    def record_ball_pickup(self, ball: Ball, position: Array) -> EventsManager:
        idx = jnp.where(ball.position == position, size=1)[0][0]
        ball = ball[idx]
        return self.replace(
            ball_pickup=Event(
                position=ball.position,
                colour=ball.colour,
                happened=jnp.asarray(True, dtype=jnp.bool_),
                event_type=EventType.PICKUP,
            )
        )


class State(struct.PyTreeNode):
    """The Markovian state of the environment"""

    key: Array
    """The random number generator state"""
    grid: Array
    """The base map of the environment that remains constant throughout the training"""
    cache: RenderingCache
    """The rendering cache to speed up rendering"""
    entities: Dict[str, Entity] = struct.field(default_factory=dict)
    """The entities in the environment, indexed via entity type string representation.
    Batched over the number of entities for each type"""
    events: EventsManager = EventsManager()
    """A struct indicating which events happened this timestep. For example, the
    goal is reached, or the player is hit by a ball."""
    mission: Event | None = None

    def get_entity(self, entity_enum: str) -> Entity:
        return self.entities[entity_enum]

    def set_entity(self, entity_enum: str, entity: Entity) -> State:
        self.entities[entity_enum] = entity
        return self

    def get_walls(self) -> Wall:
        return self.entities.get(Entities.WALL, Wall())  # type: ignore

    def set_walls(self, walls: Wall) -> State:
        self.entities[Entities.WALL] = walls
        return self

    def get_player(self, idx: int = 0) -> Player:
        return self.entities[Entities.PLAYER][idx]  # type: ignore

    def set_player(self, player: Player, idx: int = 0) -> State:
        # TODO(epignatelli): this is a hack and won't work in multi-agent settings
        self.entities[Entities.PLAYER] = player[None]
        return self

    def get_goals(self) -> Goal:
        return self.entities[Entities.GOAL]  # type: ignore

    def set_goals(self, goals: Goal) -> State:
        self.entities[Entities.GOAL] = goals
        return self

    def get_keys(self) -> Key:
        return self.entities[Entities.KEY]  # type: ignore

    def set_keys(self, keys: Key) -> State:
        self.entities[Entities.KEY] = keys
        return self

    def get_doors(self) -> Door:
        return self.entities[Entities.DOOR]  # type: ignore

    def set_doors(self, doors: Door) -> State:
        self.entities[Entities.DOOR] = doors
        return self

    def get_lavas(self) -> Lava:
        return self.entities[Entities.LAVA]  # type: ignore

    def get_balls(self) -> Ball:
        return self.entities[Entities.BALL]  # type: ignore

    def get_boxes(self) -> Ball:
        return self.entities[Entities.BOX]  # type: ignore

    def set_balls(self, balls: Ball) -> State:
        self.entities[Entities.BALL] = balls
        return self

    def set_boxes(self, boxes: Box) -> State:
        self.entities[Entities.BOX] = boxes
        return self

    def set_events(self, events: EventsManager) -> State:
        return self.replace(events=events)

    def get_positions(self) -> Array:
        return jnp.concatenate([self.entities[k].position for k in self.entities])

    def get_tags(self) -> Array:
        return jnp.concatenate([self.entities[k].tag for k in self.entities])

    def get_sprites(self) -> Array:
        return jnp.concatenate([self.entities[k].sprite for k in self.entities])

    def get_transparency(self) -> Array:
        return jnp.concatenate([self.entities[k].transparent for k in self.entities])



# Copyright 2023 The Navix Authors.

# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations
from typing import Callable


import jax.numpy as jnp
from jax import Array

from . import events
from .states import State


def compose(
    *reward_functions: Callable[[State, Array, State], Array],
    operator: Callable = jnp.sum,
) -> Callable:
    return lambda prev_state, action, state: operator(
        jnp.asarray(
            [f(prev_state, action, state) for f in reward_functions], dtype=jnp.float32
        )
    )


def free(state: State) -> Array:
    return jnp.asarray(0.0, dtype=jnp.float32)


def on_goal_reached(prev_state: State, action: Array, state: State) -> Array:
    return jnp.asarray(events.on_goal_reached(state), dtype=jnp.float32)


def action_cost(
    prev_state: State, action: Array, new_state: State, cost: float = 0.01
) -> Array:
    # noops are free
    return -jnp.asarray(action > 0, dtype=jnp.float32) * cost


def time_cost(
    prev_state: State, action: Array, new_state: State, cost: float = 0.01
) -> Array:
    # time always has a cost
    return -jnp.asarray(cost, dtype=jnp.float32)


def wall_hit_cost(
    prev_state: State, action: Array, state: State, cost: float = 0.01
) -> Array:
    return jnp.asarray(events.on_wall_hit(state), dtype=jnp.float32) * cost


def on_door_done(
    prev_state: State, action: Array, state: State, cost: float = 0.01
) -> Array:
    return jnp.asarray(events.on_door_done(state), dtype=jnp.float32)


DEFAULT_TASK = compose(on_goal_reached, action_cost)



# Copyright 2023 The Navix Authors.

# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.


from typing import Union
import jax
import jax.numpy as jnp
from jax import Array
from flax import struct

from ..components import EMPTY_POCKET_ID
from ..rendering.cache import RenderingCache
from . import Environment
from ..entities import Player, Goal, Lava
from ..states import State
from . import Timestep
from ..grid import room
from .registry import register_env


class Crossings(Environment):
    n_crossings: int = struct.field(pytree_node=False, default=1)
    is_lava: bool = struct.field(pytree_node=False, default=False)

    def reset(self, key: Array, cache: Union[RenderingCache, None] = None) -> Timestep:
        assert (
            self.height == self.width
        ), f"Crossings are only defined for square grids, got height {self.height} and \
            width {self.width}"
        # check minimum height and width
        key, k1, k2 = jax.random.split(key, num=3)

        grid = jnp.zeros((self.height - 2, self.width - 2), dtype=jnp.int32)

        # player
        player_pos = jnp.asarray([1, 1])
        player_dir = jnp.asarray(0)
        player = Player(
            position=player_pos, direction=player_dir, pocket=EMPTY_POCKET_ID
        )
        # goal
        goal_pos = jnp.asarray([self.height - 2, self.width - 2])
        goals = Goal(position=goal_pos, probability=jnp.asarray(1.0))

        entities = {
            "player": player[None],
            "goal": goals[None],
        }

        # crossings
        obstacles_hor = jnp.mgrid[
            1 : self.height - 2 : 2, 1 : self.width - 1
        ].transpose(1, 2, 0)
        obstacles_ver = jnp.mgrid[
            1 : self.height - 1, 1 : self.width - 2 : 2
        ].transpose(2, 1, 0)
        all_obstacles_pos = jnp.concatenate([obstacles_hor, obstacles_ver])
        num_obstacles = min(self.n_crossings, len(all_obstacles_pos))
        obstacles_pos = jax.random.choice(
            k1, all_obstacles_pos, (num_obstacles,), replace=False
        )

        if self.is_lava:
            entities["lava"] = Lava.create(position=obstacles_pos)
        else:
            grid = grid.at[tuple(obstacles_pos.T)].set(-1)

        # path to goal
        def update(direction, start, grid, step_size):
            return jax.lax.cond(
                direction == jnp.asarray(0, dtype=jnp.int32),
                lambda: (
                    start + jnp.asarray([0, step_size]),
                    jax.lax.dynamic_update_slice(
                        grid, jnp.zeros((1, step_size), dtype=jnp.int32), tuple(start.T)
                    ),
                ),
                lambda: (
                    start + jnp.asarray([step_size, 0]),
                    jax.lax.dynamic_update_slice(
                        grid, jnp.zeros((step_size, 1), dtype=jnp.int32), tuple(start.T)
                    ),
                ),
            )

        start = jnp.asarray([0, 0], dtype=jnp.int32)
        step_size = 3
        for i in range(10):
            k2, k3 = jax.random.split(k2)
            direction = jax.random.randint(k2, (), minval=0, maxval=2)
            start, grid = update(direction, start, grid, step_size)

        grid = jnp.pad(grid, 1, mode="constant", constant_values=-1)

        state = State(
            key=key,
            grid=grid,
            cache=RenderingCache.init(grid),
            entities=entities,
        )
        return Timestep(
            t=jnp.asarray(0, dtype=jnp.int32),
            observation=self.observation(state),
            action=jnp.asarray(-1, dtype=jnp.int32),
            reward=jnp.asarray(0.0, dtype=jnp.float32),
            step_type=jnp.asarray(0, dtype=jnp.int32),
            state=state,
        )


register_env(
    "Navix-Crossings-S9N1-v0",
    lambda *args, **kwargs: Crossings(
        *args, **kwargs, height=9, width=9, n_crossings=1
    ),
)
register_env(
    "Navix-Crossings-S9N2-v0",
    lambda *args, **kwargs: Crossings(
        *args, **kwargs, height=9, width=9, n_crossings=2
    ),
)
register_env(
    "Navix-Crossings-S9N3-v0",
    lambda *args, **kwargs: Crossings(
        *args, **kwargs, height=9, width=9, n_crossings=3
    ),
)
register_env(
    "Navix-Crossings-S11N5-v0",
    lambda *args, **kwargs: Crossings(
        *args, **kwargs, height=11, width=11, n_crossings=5
    ),
)



# Copyright 2023 The Navix Authors.

# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.


from __future__ import annotations
from typing import Union

import jax
import jax.numpy as jnp
from jax import Array
from flax import struct

from .. import rewards, terminations
from ..components import EMPTY_POCKET_ID
from ..entities import Entities, Door, Player
from ..states import EventType, State, Event
from ..grid import random_colour, random_directions
from ..rendering.cache import RenderingCache
from .environment import Environment, Timestep
from .registry import register_env


class GoToDoor(Environment):
    split_lava: bool = struct.field(pytree_node=False, default=False)

    def reset(self, key: Array, cache: Union[RenderingCache, None] = None) -> Timestep:
        # map
        grid = jnp.zeros((self.height, self.width), dtype=jnp.int32)

        k1, k2, k3, k4, k5, k6 = jax.random.split(key, num=6)
        room_height = jax.random.randint(k1, (), minval=5, maxval=self.height)
        room_width = jax.random.randint(k1, (), minval=5, maxval=self.width)

        # set wall on grid
        grid = grid.at[jnp.asarray([0, room_height - 1])].set(-1)
        grid = grid.at[:, jnp.asarray([0, room_width - 1])].set(-1)

        # player
        player_row = jax.random.randint(k2, (), minval=1, maxval=room_height - 1)
        player_col = jax.random.randint(k3, (), minval=1, maxval=room_width - 1)
        player_pos = jnp.asarray([player_row, player_col])
        direction = random_directions(k4)
        player = Player(
            position=player_pos,
            direction=direction,
            pocket=EMPTY_POCKET_ID,
        )

        # doors
        k6, k7 = jax.random.split(k5, num=2)
        rows = jax.random.randint(k6, (2,), minval=2, maxval=room_height - 2)
        cols = jax.random.randint(k7, (2,), minval=2, maxval=room_width - 2)
        positions = jnp.asarray(
            [
                [rows[0], room_width - 1],
                [room_height - 1, cols[0]],
                [rows[1], 0],
                [0, cols[1]],
            ]
        )
        colours = random_colour(key, n=4)
        open = jnp.asarray([0] * 4)
        requires = jnp.asarray([-1] * 4)
        doors = Door.create(
            position=positions, requires=requires, colour=colours, open=open
        )

        entities = {
            Entities.PLAYER: player[None],
            Entities.DOOR: doors,
        }

        idx = jax.random.randint(k6, (), minval=0, maxval=4)
        target_door = doors[idx]
        mission = Event(
            position=target_door.position,
            colour=target_door.colour,
            happened=jnp.asarray(False),
            event_type=EventType.REACH,
        )

        # systems
        state = State(
            key=key,
            grid=grid,
            cache=RenderingCache.init(grid),
            entities=entities,
            mission=mission,
        )

        return Timestep(
            t=jnp.asarray(0, dtype=jnp.int32),
            observation=self.observation(state),
            action=jnp.asarray(0, dtype=jnp.int32),
            reward=jnp.asarray(0.0, dtype=jnp.float32),
            step_type=jnp.asarray(0, dtype=jnp.int32),
            state=state,
        )


register_env(
    "Navix-GoToDoor-5x5-v0",
    lambda *args, **kwargs: GoToDoor(
        height=5,
        width=5,
        reward_fn=rewards.on_door_done,
        termination_fn=terminations.on_door_done,
        *args,
        **kwargs,
    ),
)
register_env(
    "Navix-GoToDoor-6x6-v0",
    lambda *args, **kwargs: GoToDoor(
        height=6,
        width=6,
        reward_fn=rewards.on_door_done,
        termination_fn=terminations.on_door_done,
        *args,
        **kwargs,
    ),
)
register_env(
    "Navix-GoToDoor-8x8-v0",
    lambda *args, **kwargs: GoToDoor(
        height=8,
        width=8,
        reward_fn=rewards.on_door_done,
        termination_fn=terminations.on_door_done,
        *args,
        **kwargs,
    ),
)



# Copyright 2023 The Navix Authors.

# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations
from typing import Tuple

import jax
from jax import Array
import jax.numpy as jnp
import jax.tree_util as jtu

from .entities import Entities
from .states import EventsManager, State
from .components import DISCARD_PILE_COORDS
from .grid import translate, rotate, positions_equal


class Directions:
    EAST = jnp.asarray(0)
    SOUTH = jnp.asarray(1)
    WEST = jnp.asarray(2)
    NORTH = jnp.asarray(3)


def _rotate(state: State, spin: int) -> State:
    if Entities.PLAYER not in state.entities:
        return state

    player = state.get_player(idx=0)

    # update player's direction
    direction = rotate(player.direction, spin)

    # update sprite representation
    player = player.replace(direction=direction)

    state = state.set_player(player)

    return state


def _can_walk_there(state: State, position: Array) -> Tuple[Array, EventsManager]:
    # according to the grid
    walkable = jnp.equal(state.grid[tuple(position)], 0)
    events = jax.lax.cond(
        walkable,
        lambda: state.events,
        lambda: state.events.record_grid_hit(position),
    )

    for k in state.entities:
        same_position = positions_equal(state.entities[k].position, position)
        events = jax.lax.cond(
            jnp.any(same_position),
            lambda x: x.record_walk_into(state.entities[k], position),
            lambda x: x,
            events,
        )
        obstructs = jnp.logical_and(
            jnp.logical_not(state.entities[k].walkable), same_position
        )
        walkable = jnp.logical_and(walkable, jnp.any(jnp.logical_not(obstructs)))
    return jnp.asarray(walkable, dtype=jnp.bool_), events


def _move(state: State, direction: Array) -> State:
    if Entities.PLAYER not in state.entities:
        return state

    player = state.get_player(idx=0)
    new_position = translate(player.position, direction)
    can_move, events = _can_walk_there(state, new_position)
    new_position = jnp.where(can_move, new_position, player.position)
    # update structs
    player = player.replace(position=new_position)
    state = state.set_player(player)
    return state.replace(events=events)


def noop(state: State) -> State:
    return state


def rotate_cw(state: State) -> State:
    return _rotate(state, 1)


def rotate_ccw(state: State) -> State:
    return _rotate(state, -1)


def forward(state: State) -> State:
    player = state.get_player(idx=0)
    return _move(state, player.direction)


def right(state: State) -> State:
    player = state.get_player(idx=0)
    return _move(state, player.direction + 1)


def backward(state: State) -> State:
    player = state.get_player(idx=0)
    return _move(state, player.direction + 2)


def left(state: State) -> State:
    player = state.get_player(idx=0)
    return _move(state, player.direction + 3)


def pickup(state: State) -> State:
    if Entities.KEY not in state.entities:
        return state

    player = state.get_player(idx=0)
    keys = state.get_keys()

    position_in_front = translate(player.position, player.direction)

    key_found = positions_equal(position_in_front, keys.position)

    # update keys
    positions = jnp.where(key_found, DISCARD_PILE_COORDS, keys.position)
    keys = keys.replace(position=positions)

    # update player's pocket, if the pocket has something else, we overwrite it
    key = jnp.sum(keys.id * key_found, dtype=jnp.int32)
    player = jax.lax.cond(
        jnp.any(key_found), lambda: player.replace(pocket=key), lambda: player
    )

    # update events
    events = jax.lax.cond(
        jnp.any(key_found),
        lambda: state.events.record_key_pickup(keys, position_in_front),
        lambda: state.events,
    )

    state = state.set_player(player)
    state = state.set_keys(keys)
    state = state.set_events(events)
    return state


def drop(state: State) -> State:
    raise NotImplementedError()


def toggle(state: State) -> State:
    raise NotImplementedError()


def open(state: State) -> State:
    """Unlocks and opens an openable object (like a door) if possible"""

    if Entities.DOOR not in state.entities:
        return state

    # get the tile in front of the player
    player = state.get_player(idx=0)
    doors = state.get_doors()

    position_in_front = translate(player.position, player.direction)

    # check if there is a door in front of the player
    door_found = positions_equal(position_in_front, doors.position)

    # and that, if so, either it does not require a key or the player has the key
    locked = doors.requires != -1
    key_match = player.pocket == doors.requires
    can_open = door_found & (key_match | ~locked)

    # update doors if closed and can_open
    do_open = ~doors.open & can_open
    open = jnp.where(do_open, True, doors.open)
    requires = jnp.where(do_open, -1, doors.requires)
    doors = doors.replace(open=open, requires=requires)

    # remove key from player's pocket
    pocket = jnp.asarray(player.pocket * jnp.any(can_open), dtype=jnp.int32)
    player = jax.lax.cond(
        jnp.any(can_open), lambda: player.replace(pocket=pocket), lambda: player
    )

    # update events
    events = jax.lax.cond(
        jnp.any(do_open),
        lambda: state.events.record_door_opening(doors, position_in_front),
        lambda: state.events,
    )

    state = state.set_player(player)
    state = state.set_doors(doors)
    state = state.set_events(events)

    return state


def done(state: State) -> State:
    return state


# DEFAULT_ACTION_SET = (
#     rotate_ccw,
#     rotate_cw,
#     forward,
#     pickup,
#     drop,
#     toggle,
#     done
# )
"""Default action set from Minigrid. See
https://github.com/Farama-Foundation/Minigrid/blob/master/minigrid/core/actions.py"""


COMPLETE_ACTION_SET = (
    noop,
    rotate_cw,
    rotate_ccw,
    forward,
    right,
    backward,
    left,
    pickup,
    open,
    done,
)


DEFAULT_ACTION_SET = COMPLETE_ACTION_SET
